# リスト6-4   国際線乗客数のデータからモデルを作り、予測を試みるプログラム
#%matplotlib inline  # Jupyter Notebookで実行する場合コメントマーク#を外す
import numpy as np
import pandas as pd
import math
from matplotlib import pylab as plt
from matplotlib.pylab import rcParams
rcParams['figure.figsize'] = 15, 6  # グラフを横長にする

# 統計モデルパッケージ
import statsmodels.api as sm

# https://stat.ethz.ch/R-manual/R-devel/library/datasets/html/AirPassengers.html
from rpy2.robjects import r, pandas2ri
pandas2ri.activate()
# AirPassengersデータはnumpy.arrrayの形で来る
df = pd.DataFrame(r['AirPassengers'], columns=['Airline Passengers'])
df.set_index(pd.DatetimeIndex(start='1949-01-01', end='1960-12-31', freq='M'), \
    inplace=True)

# 自己相関
print('自己相関', sm.tsa.stattools.acf(df, unbiased=True, nlags=40))
fig = sm.graphics.tsa.plot_acf(df, unbiased=True, lags=40)
fig.savefig('AirPassenger_autocorr.png')
# 偏自己相関
print('偏自己相関', sm.tsa.stattools.pacf(df, nlags=40, method='ols'))
fig2 = sm.graphics.tsa.plot_pacf(df, lags=40)
fig2.savefig('AirPassenger_pautocorr.png')

# 成分分解　季節性・トレンドを抽出
seasonal = sm.tsa.seasonal_decompose(df['Airline Passengers'].values, freq=12)
rcParams['figure.figsize'] = 15, 15
# seasonal.plot()
# plt.plot(seasonal.trend)などもできる。この場合pltにfigsizeやxlabel、titleなど設定できる
# plt.savefig('AirPassengers_decompose.png')
# plt.show()

# 対数をとる
dfl = df.apply(np.log)  # dfの全要素のnp.logをとる
# rcParams['figure.figsize'] = 15, 6
# dfl.plot(title='logged')
# plt.show()
# トレンド・季節性変動などに分解する
lseasonal = sm.tsa.seasonal_decompose(dfl['Airline Passengers'].values, freq=12)
rcParams['figure.figsize'] = 15, 15
lseasonal.plot()
plt.show()

# 定常性の確認（対数化後）
lresadf = sm.tsa.stattools.adfuller(dfl['Airline Passengers'].values)
print('p-value = {:.4}'.format(lresadf[1]))
# 出力は p-value = 0.4224　p>0.05であり、帰無仮説（定常性を持つ）が棄却されない
# ＝持つとはいえない
# 定常性がないらしいので、トレンド和分を差分をとって消す、季節性を12か月差分をとって消す
# 季節調整をしてから差分をとる
lsdiff = (dfl['Airline Passengers'] - lseasonal.seasonal).diff().dropna()
rcParams['figure.figsize'] = 15, 6
lsdiff.plot(title='lsdiff')
plt.show()
lresadfsdiff = sm.tsa.stattools.adfuller(lsdiff)
print('p-value = {:.4}'.format(lresadfsdiff[1]))
# 出力は p-value = 8.099e-09　p<0.005なので、定常性を持つだろう

# データを学習データ1949-01-01?1959-12-31とテストデータ1960-01-01?1960-12-31に分ける
ltrain = dfl[dfl.index < '1959-12-31']
ltest = dfl[dfl.index >= '1960-01-01']

# ACFとPACFよりパラメータ候補を選定  季節調整済みの差分データのコレログラム
fig, (ax1, ax2) = plt.subplots(nrows=2, sharex=True)
sm.graphics.tsa.plot_acf(lsdiff, lags=40, ax=ax1)
sm.graphics.tsa.plot_pacf(lsdiff, lags=40, ax=ax2)
plt.plot()
# acfより、pは[0,4]
# pacfより、qは[0,4]か[0,11]
# arma_order_select_icを使い、候補から当てはまりのよいモデルのパラメータを選ぶ
# これだとp=4, q=4が最適
lresDiff = sm.tsa.arma_order_select_ic(lsdiff, ic='aic', trend='nc', \
    max_ar=4, max_ma=4)
print(lresDiff)
print(lresDiff.aic_min_order)
# これだとp=2, q=2が最適

# SARIMAX適用
from statsmodels.tsa.statespace.sarimax import SARIMAX
lmodel = SARIMAX(
    ltrain,
    order=(2, 1, 2),  # p,d,q = 2,1,2
    seasonal_order=(1, 1, 1, 12),  # P,D,Q,s = 1,1,1,12
    trend='t',
    enforce_stationarity=False,
    enforce_invertibility=False)
lresult = lmodel.fit()
# 予測
ltrain_pred = lresult.predict()
ltest_pred = lresult.forecast(len(ltest))
ltest_pred_ci = lresult.get_forecast(len(ltest)).conf_int()
# print('dfl\n', dfl['Airline Passengers'])
# print('ltrain\n', ltrain)
# print('ltest_pred\n', ltest_pred)
# print(ltest_pred_ci)

# 予測の評価
from sklearn.metrics import mean_squared_error
ltrain_rmse = np.sqrt(mean_squared_error(ltrain, ltrain_pred))
ltest_rmse = np.sqrt(mean_squared_error(ltest, ltest_pred))
print('RMSE(train): {:.5}\nRMSE(ltest): {:.5}'.format(ltrain_rmse, ltest_rmse))

# 結果のプロット
fig, ax = plt.subplots()
dfl['Airline Passengers'].plot(ax=ax, label='Original', linestyle="dashed")
ltrain_pred.plot(ax=ax, label='Predict(train)')
ltest_pred.plot(ax=ax, label='Predict(test)')
ax.fill_between(
    ltest_pred_ci.index,
    ltest_pred_ci.iloc[:, 0],
    ltest_pred_ci.iloc[:, 1],
    color='k',
    alpha=.2)  # 5%信頼区間をグレーで塗る
ax.legend()
# ax.legend(loc='lower right')  # 凡例の位置を右上に移したい場合
plt.title('model result')
plt.show()
